import uuid
from datetime import datetime
from typing import Dict, Any, Optional
from uuid import UUID

from pydantic import ConfigDict, BaseModel

from ul_unipipeline.answer.uni_answer_params import UniAnswerParams
from ul_unipipeline.message_meta.uni_message_meta_err import UniMessageMetaErr
from ul_unipipeline.message_meta.uni_message_meta_error_topic import UniMessageMetaErrTopic


class UniMessageMeta(BaseModel):
    id: UUID
    date_created: datetime
    payload: Dict[str, Any]

    parent: Optional[Dict[str, Any]] = None
    error: Optional[UniMessageMetaErr] = None

    unwrapped: bool
    answer_params: Optional[UniAnswerParams] = None

    worker_creator: Optional[str] = None  # name of worker who created it

    ttl_s: Optional[int] = None

    model_config = ConfigDict(
        frozen=True,
        extra="ignore",
    )

    @property
    def real_ttl_s(self) -> Optional[int]:
        return self.ttl_s or (self.answer_params.ttl_s if self.answer_params is not None else None)

    @property
    def need_answer(self) -> bool:
        return self.answer_params is not None

    @property
    def has_error(self) -> bool:
        return self.error is not None

    @staticmethod
    def create_new(data: Dict[str, Any], unwrapped: bool, answer_params: Optional[UniAnswerParams] = None, error: Optional[UniMessageMetaErr] = None, ttl_s: Optional[int] = None) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=uuid.uuid4(),
            date_created=datetime.now(),
            payload=data,
            parent=None,
            error=error,
            answer_params=answer_params,
            unwrapped=unwrapped,
            ttl_s=ttl_s,
        )

    def create_child(self, payload: Dict[str, Any], unwrapped: bool, answer_params: Optional[UniAnswerParams] = None, ttl_s: Optional[int] = None) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=uuid.uuid4(),
            date_created=datetime.now(),
            payload=payload,
            parent=self.model_dump(),
            unwrapped=unwrapped,
            answer_params=answer_params,
            error=None,
            ttl_s=ttl_s,
        )

    def create_error_child(self, error_topic: UniMessageMetaErrTopic, error: Exception, ttl_s: Optional[int] = None) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=self.id,
            date_created=self.date_created,
            payload=self.payload,
            parent=self.model_dump(),
            unwrapped=self.unwrapped,
            answer_params=self.answer_params,
            error=UniMessageMetaErr(
                error_topic=error_topic,
                error_type=type(error).__name__,
                error_message=str(error),
                retry_times=self.error.retry_times + 1 if self.error is not None else 0
            ),
            ttl_s=ttl_s,
        )



from typing import Set, Optional
from uuid import UUID

from ul_unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from ul_unipipeline.definitions.uni_definition import UniDefinition
from ul_unipipeline.definitions.uni_message_definition import UniMessageDefinition
from ul_unipipeline.definitions.uni_module_definition import UniModuleDefinition
from ul_unipipeline.definitions.uni_waiting_definition import UniWaitingDefinition


class UniWorkerDefinition(UniDefinition):
    id: UUID
    name: str
    broker: UniBrokerDefinition
    type: Optional[UniModuleDefinition] = None
    topic: str
    prefetch_count: int
    error_topic: str
    error_payload_topic: str
    answer_topic: str
    input_message: UniMessageDefinition
    answer_message: Optional[UniMessageDefinition] = None
    output_workers: Set[str]
    waitings: Set[UniWaitingDefinition]
    external: Optional[str] = None
    input_unwrapped: bool
    answer_unwrapped: bool

    @property
    def marked_as_external(self) -> bool:
        return self.external is not None

    @property
    def need_answer(self) -> bool:
        return self.answer_message is not None



from ul_unipipeline.definitions.uni_definition import UniDefinition
from ul_unipipeline.definitions.uni_module_definition import UniModuleDefinition


class UniMessageDefinition(UniDefinition):
    name: str
    type: UniModuleDefinition



from typing import Optional
from uuid import UUID

from pydantic import model_validator, field_validator
from pydantic_core.core_schema import ValidationInfo

from ul_unipipeline.definitions.uni_definition import UniDefinition
from ul_unipipeline.definitions.uni_worker_definition import UniWorkerDefinition


class UniCronTaskDefinition(UniDefinition):
    id: UUID
    name: str
    worker: UniWorkerDefinition
    when: Optional[str] = None
    every_sec: Optional[int] = None
    alone: bool

    @field_validator('every_sec')
    @classmethod
    def validate_every_sec(cls, v: Optional[int], info: ValidationInfo) -> Optional[int]:
        if v is None:
            return v
        if not isinstance(v, int):
            TypeError('Invalid type. must be int')
        if v <= 1:
            raise ValueError('Must be > 1')
        if 60 % v:
            raise ValueError('60 must be a multiple of N')
        return v

    @model_validator(mode='after')
    def validate_all(self) -> 'UniCronTaskDefinition':
        when = self.when
        every_sec = self.every_sec
        if not ((when is not None) ^ (every_sec is not None)):
            raise ValueError(f'cron "{self.name}" has property conflict in (when, every_sec). one from it must be None')
        return self



from pydantic import ConfigDict, BaseModel

from ul_unipipeline.message_meta.uni_message_meta_error_topic import UniMessageMetaErrTopic


class UniMessageMetaErr(BaseModel):
    error_topic: UniMessageMetaErrTopic
    error_type: str
    error_message: str
    retry_times: int

    model_config = ConfigDict(
        extra="ignore",
        frozen=True,
    )



from uuid import UUID

from pydantic import ConfigDict, BaseModel


class UniAnswerParams(BaseModel):
    topic: str
    id: UUID
    ttl_s: int

    model_config = ConfigDict(
        frozen=True,
        extra="ignore",
    )



from typing import Optional


class UniError(Exception):
    pass


class UniAnswerDelayError(UniError):
    pass


class UniRedundantAnswerError(UniError):
    pass


class UniEmptyAnswerError(UniError):
    pass


class UniConfigError(UniError):
    pass


class UniDefinitionNotFoundError(UniError):
    pass


class UniMessageError(UniError):
    pass


class UniMessagePayloadParsingError(UniMessageError):
    pass


class UniAnswerMessagePayloadParsingError(UniMessageError):
    pass


class UniPayloadSerializationError(UniMessageError):
    pass


class UniSendingToUndefinedWorkerError(UniError):
    pass


class UniTopicNotFoundError(UniError):
    pass


class UniMessageRejectError(UniError):

    def __init__(self, exc: Optional[Exception] = None) -> None:
        self._exc = exc

    @property
    def rejection_exception(self) -> Optional[Exception]:
        return self._exc
